#pragma once

#include <stdint.h>

G_BEGIN_DECLS

#define PHOC_XCURSOR_SIZE 24

#define PHOC_XCURSOR_DEFAULT "left_ptr"
#define PHOC_XCURSOR_MOVE "grabbing"
#define PHOC_XCURSOR_ROTATE "grabbing"

G_END_DECLS
