/*
 * Copyright (C) 2022 Purism SPC
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "animatable.h"
#include "phoc-anim-enums.h"
#include "phoc-anim-enum-types.h"
#include "property-easer.h"
#include "timed-animation.h"

